﻿// 補間モードを設定できるようにする。
// 違いが見られなかったのでとりあえず無効化。
//#define USE_INTERPOLATION

using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;
using System.Runtime.InteropServices;
using NiVE2.Plugin.Interface;
using NiVE2.Plugin.Property;
using NiVE2.Plugin.Controls;
using NiVE2.Drawing;
using NiVE2.Utils;
using NiVE2.Utils.Cache;
using NiVE2.Utils.Threading;

namespace ruche.nive2.effects
{
    public class Text : EffectBase
    {
        /// <summary>
        /// 縦横基準に対応するアラインメント配列。
        /// </summary>
        private static readonly StringAlignment[] Alignments =
            new StringAlignment[]
                {
                    StringAlignment.Near,
                    StringAlignment.Center,
                    StringAlignment.Far,
                };

        /// <summary>
        /// 複合品質レベル名配列。
        /// </summary>
        private static readonly string[] CompositingQualityNames =
            new string[]
                {
                    "デフォルト",
                    "低品質",
                    "高品質",
                    "線形",
                    "ガンマ補正",
                };

        /// <summary>
        /// 複合品質レベル配列。
        /// </summary>
        private static readonly CompositingQuality[] CompositingQualities =
            new CompositingQuality[]
                {
                    CompositingQuality.Default,
                    CompositingQuality.HighSpeed,
                    CompositingQuality.HighQuality,
                    CompositingQuality.AssumeLinear,
                    CompositingQuality.GammaCorrected,
                };

#if USE_INTERPOLATION
        /// <summary>
        /// 補間モード名配列。
        /// </summary>
        private static readonly string[] InterpolationModeNames =
            new string[]
                {
                    "デフォルト",
                    "低品質",
                    "高品質",
                    "最近傍",
                    "双一次",
                    "高品質双一次",
                    "双三次",
                    "高品質双三次",
                };

        /// <summary>
        /// 補間モード配列。
        /// </summary>
        private static readonly InterpolationMode[] InterpolationModes =
            new InterpolationMode[]
                {
                    InterpolationMode.Default,
                    InterpolationMode.Low,
                    InterpolationMode.High,
                    InterpolationMode.NearestNeighbor,
                    InterpolationMode.Bilinear,
                    InterpolationMode.HighQualityBilinear,
                    InterpolationMode.Bicubic,
                    InterpolationMode.HighQualityBicubic,
                };
#endif // USE_INTERPOLATION

        /// <summary>
        /// 文字列描画フォーマット。
        /// </summary>
        private static readonly StringFormat BaseStringFormat =
            new StringFormat(
                StringFormatFlags.DisplayFormatControl  |
                StringFormatFlags.MeasureTrailingSpaces |
                StringFormatFlags.NoClip                |
                StringFormatFlags.NoWrap);

        /// <summary>
        /// 対象レイヤ。
        /// </summary>
        private ILayer _layer = null;

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        public Text()
        {
        }

        /// <summary>
        /// ブラシを作成する。
        /// </summary>
        /// <param name="baseColor">基本色。</param>
        /// <param name="alpha">不透明度。 0 ～ 255 。</param>
        /// <param name="gradation">
        /// グラデーションするならば true 。しないならば false 。
        /// </param>
        /// <param name="gradationBaseY">グラデーションの基準Y位置。</param>
        /// <param name="gradationHeight">グラデーションの高さ。</param>
        /// <param name="gradationColor">グラデーションの終端色。</param>
        /// <returns>ブラシ。</returns>
        private static Brush CreateBrush(
            Color baseColor,
            int alpha,
            bool gradation,
            float gradationBaseY,
            float gradationHeight,
            Color gradationColor)
        {
            Brush brush = null;

            baseColor = Color.FromArgb(alpha, baseColor);

            if (gradation)
            {
                gradationColor = Color.FromArgb(alpha, gradationColor);
                brush = new LinearGradientBrush(
                    new PointF(0.0F, gradationBaseY),
                    new PointF(0.0F, gradationBaseY + gradationHeight),
                    baseColor,
                    gradationColor);
            }
            else
            {
                brush = new SolidBrush(baseColor);
            }

            return brush;
        }

        #region EffectBase メンバ

        public override string Category
        {
            get { return "テキスト"; }
        }

        public override bool IsAudioEffect
        {
            get { return false; }
        }

        public override bool IsVideoEffect
        {
            get { return true; }
        }

        public override void Initialize(ILayer layer)
        {
            _layer = layer;
        }

        public override object SaveInnerData()
        {
            return null;
        }

        public override void LoadInnerData(object data)
        {
        }

        public override PropertyBase[] GetDefaultProperty()
        {
            return new PropertyBase[]
                {
                    new StringProperty("テキスト", ""),
                    new FontProperty("フォント", new Font(FontFamily.GenericSansSerif, 14.0F)),
                    new NumberProperty("不透明度", 100.0, 100.0, 0.0),
                    // 配置
                        new VertexProperty("配:基準点", Point.Empty, double.MaxValue, double.MinValue),
                        new SelectableProperty("配:横基準", 0, new string[]{ "左", "中央", "右" }),
                        new SelectableProperty("配:縦基準", 0, new string[]{ "上", "中央", "下" }),
                        new NumberProperty("配:行間", 100.0, 1000.0, 0.0),
                    // 塗り潰し
                        new BooleanProperty("塗:有効", true),
                        new ColorProperty("塗:色", Color.Black),
                        // グラデーション
                            new BooleanProperty("塗:グ:有効", false),
                            new ColorProperty("塗:グ:色", Color.Black),
                    // 縁取り
                        new BooleanProperty("縁:有効", false),
                        new NumberProperty("縁:太さ", 1.0, double.MaxValue, 0.0),
                        new ColorProperty("縁:色", Color.White),
                        // グラデーション
                            new BooleanProperty("縁:グ:有効", false),
                            new ColorProperty("縁:グ:色", Color.Black),
                    // 品質
                        new BooleanProperty("品:アンチエイリアス", true),
                        new SelectableProperty("品:複合品質", 0, CompositingQualityNames),
#if USE_INTERPOLATION
                        new SelectableProperty("品:補間モード", 0, InterpolationModeNames),
#endif // USE_INTERPOLATION
                };
        }

        public override PropertyEditControlBase[] GetControl()
        {
            return new PropertyEditControlBase[]
                {
                    new StringPropertyEditControl("テキスト"),
                    new FontPropertyEditControl("フォント"),
                    new NumberPropertyEditControl("不透明度", 1.0, NumberPropertyEditControlType.Double),
                    new PropertyNestBeginControl("配置"),
                        new VertexPropertyEditControl("配:基準点", 1.0, VertexPropertyEditControlType.PointFOrSizeF),
                        new SelectablePropertyEditControl("配:横基準"),
                        new SelectablePropertyEditControl("配:縦基準"),
                        new NumberPropertyEditControl("配:行間", 1.0, NumberPropertyEditControlType.Double),
                    new PropertyNestEndControl(),
                    new PropertyNestBeginControl("塗り潰し"),
                        new BooleanPropertyEditControl("塗:有効"),
                        new ColorPropertyEditControl("塗:色"),
                        new PropertyNestBeginControl("グラデーション"),
                            new BooleanPropertyEditControl("塗:グ:有効"),
                            new ColorPropertyEditControl("塗:グ:色"),
                        new PropertyNestEndControl(),
                    new PropertyNestEndControl(),
                    new PropertyNestBeginControl("縁取り"),
                        new BooleanPropertyEditControl("縁:有効"),
                        new NumberPropertyEditControl("縁:太さ", 1.0, NumberPropertyEditControlType.Double),
                        new ColorPropertyEditControl("縁:色"),
                        new PropertyNestBeginControl("グラデーション"),
                            new BooleanPropertyEditControl("縁:グ:有効"),
                            new ColorPropertyEditControl("縁:グ:色"),
                        new PropertyNestEndControl(),
                    new PropertyNestEndControl(),
                    new PropertyNestBeginControl("品質"),
                        new BooleanPropertyEditControl("品:アンチエイリアス"),
                        new SelectablePropertyEditControl("品:複合品質"),
#if USE_INTERPOLATION
                        new SelectablePropertyEditControl("品:補間モード"),
#endif // USE_INTERPOLATION
                    new PropertyNestEndControl(),
                };
        }

        public override Roi CheckRoi(
            Roi roi,
            ReadOnlyDictionary<string, PropertyBase> property,
            double time)
        {
            // 特に変更しない
            return roi;
        }

        public override NBitmap ProcessingImage(
            NBitmap image,
            ReadOnlyDictionary<string, PropertyBase> property,
            Roi roi,
            double time)
        {
            // プロパティ取得
            string text = ((StringProperty)property["テキスト"]).Text;
            FontProperty fontProp = (FontProperty)property["フォント"];
            double alpha = ((NumberProperty)property["不透明度"]).DoubleValue;
            PointF basePos = ((VertexProperty)property["配:基準点"]).PointFValue;
            int horzIndex = ((SelectableProperty)property["配:横基準"]).Selected;
            int vertIndex = ((SelectableProperty)property["配:縦基準"]).Selected;
            double heightRate = ((NumberProperty)property["配:行間"]).DoubleValue;
            bool fillEnable = ((BooleanProperty)property["塗:有効"]).Boolean;
            Color fillColor = ((ColorProperty)property["塗:色"]).Color;
            bool fillGradEnable = ((BooleanProperty)property["塗:グ:有効"]).Boolean;
            Color fillGradColor = ((ColorProperty)property["塗:グ:色"]).Color;
            bool drawEnable = ((BooleanProperty)property["縁:有効"]).Boolean;
            double drawWidth = ((NumberProperty)property["縁:太さ"]).DoubleValue;
            Color drawColor = ((ColorProperty)property["縁:色"]).Color;
            bool drawGradEnable = ((BooleanProperty)property["縁:グ:有効"]).Boolean;
            Color drawGradColor = ((ColorProperty)property["縁:グ:色"]).Color;
            bool antiAlias = ((BooleanProperty)property["品:アンチエイリアス"]).Boolean;
            int compoIndex = ((SelectableProperty)property["品:複合品質"]).Selected;
#if USE_INTERPOLATION
            int interIndex = ((SelectableProperty)property["品:補間モード"]).Selected;
#endif // USE_INTERPOLATION

            // 描画するものが無いならそのまま返す
            if (
                text.Trim() == ""   ||
                alpha <= 0.0        ||
                (!fillEnable && !drawEnable))
            {
                return image;
            }

            // 文字列を行単位に分割
            // 行間設定、および末尾の改行を描画しないため
            List<string> lines = new List<string>();
            StringReader sr = new StringReader(text);
            string line;
            while ((line = sr.ReadLine()) != null)
            {
                lines.Add(line);
            }

            // フォントを作成(ピクセル単位)
            Font font = new Font(
                fontProp.FontFamily,
                fontProp.Size,
                fontProp.Style,
                GraphicsUnit.Pixel);

            // 高さ算出
            float fontHeight = font.Height;
            double lineHeight = fontHeight * heightRate / 100.0;
            double allHeight =
                lineHeight * lines.Count - (lineHeight - fontHeight); // 最後の余白は含まない

            // 文字列描画フォーマットを作成
            StringAlignment horzAlign = Alignments[horzIndex];
            StringAlignment vertAlign = Alignments[vertIndex];
            StringFormat format = BaseStringFormat.Clone() as StringFormat;
            format.Alignment = horzAlign;
            //format.LineAlignment = vertAlign; // 縦方向の配置は自前で行う

            // 描画開始Y位置算出
            double beginY = basePos.Y;
            switch (vertAlign)
            {
                case StringAlignment.Center:
                    beginY -= allHeight * 0.5;
                    break;

                case StringAlignment.Far:
                    beginY -= allHeight;
                    break;
            }

            // 不透明度を 0 ～ 255 にする
            int alphaVal = (int)(alpha * 255.0 / 100.0 + 0.5);

            // Bitmapを取得
            CacheManager.Manager.DemandTemporalMemory(image.DataSize);
            using (Bitmap bmp = image.ToBitmap())
            {
                CacheManager.Manager.AddWeakCache(bmp, image.DataSize);

                // 描画
                using (Graphics g = Graphics.FromImage(bmp))
                {
                    // 描画設定
                    g.SmoothingMode = antiAlias ?
                        SmoothingMode.AntiAlias : SmoothingMode.Default;
                    g.TextRenderingHint = antiAlias ?
                        TextRenderingHint.AntiAlias : TextRenderingHint.SystemDefault;
                    g.CompositingMode = CompositingMode.SourceOver;
                    g.CompositingQuality = CompositingQualities[compoIndex];
#if USE_INTERPOLATION
                    g.InterpolationMode = InterpolationModes[interIndex];
#endif // USE_INTERPOLATION

                    // ROIでクリッピング
                    g.SetClip(roi.Interest);

                    // スケールマトリクス設定
                    float scaleRate = (float)_layer.ResolutionRate;
                    g.ScaleTransform(scaleRate, scaleRate, MatrixOrder.Append);

                    // 以下の理由により、行ごとに描画する
                    // - 全行一度にパス追加すると改行文字も描画されてしまうため
                    // - グラデーションを綺麗に出すため
                    for (int i = 0; i < lines.Count; ++i)
                    {
                        line = lines[i];
                        if (line.Trim() == "") { continue; }

                        double y = beginY + (lineHeight * i);

                        using (GraphicsPath gpath = new GraphicsPath())
                        {
                            // パス追加
                            gpath.AddString(
                                line,
                                font.FontFamily,
                                (int)font.Style,
                                font.SizeInPoints * g.DpiY / 72.0F,
                                new PointF(basePos.X, (float)y),
                                format);

                            // 縁取り
                            if (drawEnable)
                            {
                                // ブラシからペン作成
                                using (
                                    Brush brush = CreateBrush(
                                        drawColor,
                                        alphaVal,
                                        drawGradEnable,
                                        (float)y,
                                        (float)fontHeight,
                                        drawGradColor))
                                using (
                                    Pen pen = new Pen(brush, (float)drawWidth))
                                {
                                    // 描画
                                    g.DrawPath(pen, gpath);
                                }
                            }

                            // 塗り潰し
                            if (fillEnable)
                            {
                                // ブラシ作成
                                using (
                                    Brush brush = CreateBrush(
                                        fillColor,
                                        alphaVal,
                                        fillGradEnable,
                                        (float)y,
                                        (float)fontHeight,
                                        fillGradColor))
                                {
                                    // 描画
                                    g.FillPath(brush, gpath);
                                }
                            }
                        }
                    }
                }

                // 元画像を上書き
                byte[] dest = image.GetData();
                BitmapData bmpData = bmp.LockBits(
                    new Rectangle(Point.Empty, bmp.Size),
                    ImageLockMode.ReadOnly,
                    PixelFormat.Format32bppArgb);
                Marshal.Copy(bmpData.Scan0, dest, 0, dest.Length);
                bmp.UnlockBits(bmpData);
                bmpData = null;
            }

            return image;
        }

        public override byte[] ProcessingAudio(
            byte[] audio,
            ReadOnlyDictionary<string, PropertyBase>[] property,
            double time)
        {
            // 呼ばれないはず
            throw new NotImplementedException();
        }

        #endregion

        #region PluginBase メンバ

        public override string PluginName
        {
            get { return "テキスト(詳細)"; }
        }

        public override string Author
        {
            get { return "ルーチェ"; }
        }

        public override string InfoLink
        {
            get { return "http://www.ruche-home.net/"; }
        }

        public override string Description
        {
            get
            {
                return "パラメータを細かく指定してテキストを描画します。";
            }
        }

        #endregion

        #region IDisposable メンバ

        public override void Dispose()
        {
            // 何もしない
        }

        #endregion
    }
}
